"""The class for testing models and reporting results

    A collection of methods to simplify your code.
"""
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import roc_auc_score
from sklearn.metrics import matthews_corrcoef
#%pip install mlxtend --upgrade
from mlxtend.evaluate import bias_variance_decomp
import seaborn as sns
import numpy as np
import pandas as pd
import pickle

class Metrics():
    def split_tuples(self, tuples):
        """
            Splits tuples of sample and its target in the relative lists

            Arguments:
                :tuples (list[tuple]): list of tuples with sample and its target
            Returns:
                tuple of list of samples and list of targets
        """
        X_test = []
        y_test = []
        for sample, target in tuples:
            X_test.append(sample)
            y_test.append(target)
        return X_test, y_test

    def prediction(self, model, method, X_test, y_test = []):
        """
            Predicts with your model

            Arguments:
                :model (obj): object of your model
                :method (str): name of method
                :X_test (list[]|list[tuple]): list of samples or list of tuples with sample and its target
                :y_test (list[]): list of targets
            Returns:
                tuple of list of targets and list of predictions
        """
        y_pred = []
        if len(X_test) > 0 and (
            (type(X_test) is pd.core.series.Series and type(X_test[X_test.first_valid_index()]) is tuple)
            or (type(X_test[:1][0]) is tuple)
        ):
            X_test, y_test = self.split_tuples(X_test)
        predictor = getattr(model, method)
        for sample in X_test:
            y_pred.append(predictor(sample))
        return y_test, y_pred

    def create_confusion_matrix(self, y_test, y_pred, is_test = False):
        """
            Creates and prints confusion matrix

            Arguments:
                :y_test (list[]): list of targets
                :y_pred (list[]): list of predictions
                :is_test (bool): default is False
            Returns:
                confusion matrix
        """
        matrix = confusion_matrix(y_test, y_pred)
        tick_labels = np.unique(np.array([y_test, y_pred]))
        if is_test == False:
            sns.heatmap(matrix, square=True, annot=True, fmt='d', cbar=False, xticklabels = tick_labels, yticklabels = tick_labels)
        return matrix

    def is_binary_classification(self, y_test, y_pred):
        """
            Gets if the classification is binary or not

            Arguments:
                :y_test (list[]): list of targets
                :y_pred (list[]): list of predictions
            Returns:
                boolean
        """
        classes = np.unique(np.array([y_test, y_pred]))
        if len(classes) == 2:
            return True
        return False

    def clean_binary_classification(self, y_test, y_pred):
        """
            Transforms the target and prediction in integer 0 and 1

            Arguments:
                :y_test (list[]): list of targets
                :y_pred (list[]): list of predictions
            Returns:
                y_test and y_pred with only 0 and 1 values
        """
        classes = np.unique(np.array([y_test, y_pred]))
        y_test = [0 if x == classes[0] else 1 for x in y_test]
        y_pred = [0 if x == classes[0] else 1 for x in y_pred]
        return y_test, y_pred

    def get_classification_metrics(self, params = {}):
        """
            Gets classification metrics

            Arguments: params (dict) with the keys below
                :model (obj): object of your model
                :X_train (list[]|list[tuple]): list of samples or list of tuples with sample and its target
                :y_train (list[]): list of targets
                :X_test (list[] | list[tuple]): list of samples or list of tuples with sample and its target
                :y_test (list[]): list of targets
                :y_pred (list[]): list of predictions
                :loss (str): parameter of bias_variance_decomp, default mse
                :num_rounds (int): parameter of bias_variance_decomp, default 200
                :random_seed (int): parameter of bias_variance_decomp, default 3
            Returns:
                dictionary with Loss, Bias, Variance, MCC, ROC_AUC, Accuracy, Precision, Recall, Fscore
        """
        loss, bias, variance, roc_auc = (0, 0, 0, 0)
        if 'model' in params:
            fit_exists = hasattr(params['model'], 'fit') and callable(getattr(params['model'], 'fit'))
            predict_exists = hasattr(params['model'], 'predict') and callable(getattr(params['model'], 'predict'))
            if fit_exists == True or predict_exists == True:
                if 'loss' not in params:
                    params['loss'] = 'mse'
                if 'num_rounds' not in params:
                    params['num_rounds'] = 200
                if 'random_seed' not in params:
                    params['random_seed'] = 3
                loss, bias, variance = bias_variance_decomp(params['model'], params['X_train'], params['y_train'], params['X_test'], params['y_test'], loss=params['loss'], num_rounds=params['num_rounds'], random_seed=params['random_seed'])
        accuracy = accuracy_score(params['y_test'], params['y_pred'])
        report = precision_recall_fscore_support(params['y_test'], params['y_pred'])
        mcc = matthews_corrcoef(params['y_test'], params['y_pred'])
        if self.is_binary_classification(params['y_test'], params['y_pred']):
            y_test, y_pred = self.clean_binary_classification(params['y_test'], params['y_pred'])
            roc_auc = roc_auc_score(y_test, y_pred)
        return {'Loss': loss, 'Bias': bias, 'Variance': variance, 'MCC': mcc, 'ROC_AUC': roc_auc, 'Accuracy': accuracy, 'Precision': report[0], 'Recall': report[1], 'Fscore': report[2], 'Support': report[3]}

    def print_metrics(self, metrics):
        """
            Prints metrics

            Arguments:
                :metrics (dict): dictionary of metrics with their value
            Returns:
                only the print of metrics
        """
        for metric in metrics.keys():
            print(metric, metrics[metric])

    def save_model(self, model, filename):
        """
            Saves model

            Arguments:
                :model (obj): object of your model
                :filename (str): pathname and filename where you want to save your model
        """
        with open(filename, 'wb') as pickle_file:
            pickle.dump(model, pickle_file)

    def resume_model(self, filename):
        """
            Resumes model

            Arguments:
                :filename (str): pathname and filename where you want to save your model
            Returns:
                object of your model
        """
        with open(filename, 'rb') as pickle_file:
          return pickle.load(pickle_file)



"""The classes for data preparation with nltk, sklearn and more

    A collection of methods to simplify your code.
"""
import re
import string
import nltk
nltk.download('punkt')
nltk.download('wordnet')
nltk.download('omw-1.4')
nltk.download('stopwords')
nltk.download('averaged_perceptron_tagger')
nltk.download('vader_lexicon')
from nltk.corpus import stopwords
from nltk.corpus import wordnet as wn
from nltk.stem import WordNetLemmatizer
from nltk.util import ngrams as ng
from collections import defaultdict
from collections import Counter
from wordcloud import WordCloud
import matplotlib.pyplot as plt
#%matplotlib inline
from sklearn.feature_extraction import DictVectorizer
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.feature_extraction.text import TfidfVectorizer
from nltk.sentiment import SentimentIntensityAnalyzer

class Ntk():
    """
    The class Ntk contains the Natural Language Processing tool kit.

    Arguments: params (dict) with the keys below
        :language (str): default is english
        :lemmatizer (obj): obj with method lemmatize() like WordNetLemmatizer()
        :min_length (int): default is 1, so all words will be used
        :stop_words (list[str]): default is stopwords.words()
        :tag_map (dict): default contains J, V, R

    Here's an example:

        >>> from smltk.preprocessing import Ntk
        >>> doc = 'Good case, Excellent value.'
        >>> ntk = Ntk()
        >>> get_doc_cleaned = ntk.get_doc_cleaned(doc)
        >>> print(get_doc_cleaned)
        good case excellent value
    """
    language = 'english'
    lemmatizer = None
    min_length = 1
    stop_words = []
    tag_map = {}
    sia = None
    vectorizer = None

    def __init__(self, params = []):
        if 'language' in params:
            self.language = params['language']
        if 'lemmatizer' in params:
            self.lemmatizer = params['lemmatizer']
        else:
            self.lemmatizer = WordNetLemmatizer()
        if 'min_length' in params:
            self.min_length = params['min_length']
        if 'stop_words' in params:
            self.stop_words = set(params['stop_words'])
        else:
            self.stop_words = set(stopwords.words(self.language))
        if 'tag_map' in params:
            self.tag_map = params['tag_map']
        else:
            self.tag_map = defaultdict(lambda : wn.NOUN)
            self.tag_map['J'] = wn.ADJ
            self.tag_map['V'] = wn.VERB
            self.tag_map['R'] = wn.ADV
        if not len(self.tag_map):
            for pos in ['E', 'D', 'J', 'N', 'V', 'R']:
                self.tag_map[pos] = wn.NOUN
        self.sia = SentimentIntensityAnalyzer()

    def word_tokenize(self, doc):
        """
            Splits document in each word

            Arguments:
                :doc (str): text
            Returns:
                list of words
        """
        return nltk.word_tokenize(doc)

    def tokenize_and_clean_doc(self, doc):
        """
            Tokenizes doc
            and filters tokens from punctuation, numbers, stop words, and words <= min_length
            and changes upper characters in lower characters

            Arguments:
                :doc (str): text
            Returns:
                list of words filtered
        """
        # split into tokens by white space
        tokens = self.word_tokenize(doc)
        # prepare regex for char filtering
        punctuation = re.compile('[%s]' % re.escape(string.punctuation))
        # remove punctuation from each word
        tokens = [punctuation.sub('', token) for token in tokens]
        # remove remaining tokens with no alphabetic characters
        tokens = [token for token in tokens if token.isalpha()]
        # convert in lower case
        tokens = [token.lower() for token in tokens]
        # filter out stop words
        tokens = [token for token in tokens if not token in self.stop_words]
        # filter out short tokens
        tokens = [token for token in tokens if len(token) >= self.min_length]
        return tokens

    def lemmatize(self, tokens):
        """
            Lemmatizes tokens

            Arguments:
                :tokens (list[str]): list of words
            Returns:
                list of tokens lemmatized
        """
        # lemmatization
        lemmas = [self.lemmatizer.lemmatize(token, self.tag_map[tag[0]]) for token, tag in nltk.pos_tag(tokens)]
        return lemmas

    def get_tokens_cleaned(self, doc, is_lemma = True):
        """
            Tokenizes doc
            and filters tokens from punctuation, numbers, stop words, and words <= min_length
            and changes upper characters in lower characters
            and if is_lemma == True, also it lemmatizes

            Arguments:
                :doc (str): text
                :is_lemma (bool): default is True
            Returns:
                list of tokens cleaned
        """
        tokens = self.tokenize_and_clean_doc(doc)
        if is_lemma == True:
            tokens = self.lemmatize(tokens)
        return tokens

    def __call__(self, doc):
        """
            Alias of get_tokens_cleaned for using in models

            Arguments:
                :doc (str): text
            Returns:
                list of tokens cleaned
        """
        return self.get_tokens_cleaned(doc)

    def get_doc_cleaned(self, doc, is_lemma = True):
        """
            Filters doc from punctuation, numbers, stop words, and words <= min_length
            and changes upper characters in lower characters
            and if is_lemma == True, also it lemmatizes

            Arguments:
                :doc (str): text
                :is_lemma (bool): default is True
            Returns:
                string cleaned
        """
        tokens = self.get_tokens_cleaned(doc, is_lemma)
        return ' '.join(map(str, tokens))

    def add_doc_to_vocab(self, doc, vocab, is_lemma = True):
        """
            Adds tokens of that doc to vocabulary and updates vocabulary

            Arguments:
                :doc (str): text
                :vocab (collections.Counter): dictionary of tokens with its count
                :is_lemma (bool): default is True
            Returns:
                list of tokens of that doc and vocab updated
        """
        tokens = self.get_tokens_cleaned(doc, is_lemma)
        vocab.update(tokens)
        return tokens

    def create_vocab_from_docs(self, docs, is_lemma = True):
        """
            Creates vocabulary from list of docs

            Arguments:
                :docs (list[str]): list of texts
                :is_lemma (bool): default is True
            Returns:
                dictionary of tokens with its count in an object collections.Counter
        """
        vocab = Counter()
        [self.add_doc_to_vocab(doc, vocab, is_lemma) for doc in docs]
        return vocab

    def get_stats_vocab(self, vocab, min_occurance = 1):
        """
            Gets statistics of vocabulary

            Arguments:
                :vocab (collections.Counter): dictionary of tokens with its count
                :min_occurance (int): minimum occurance considered
            Returns:
                tuple of tokens number with >= min_occurance and total tokens number
        """
        tokens = [ key for key, count in vocab.items() if count >= min_occurance]
        partial = len(tokens)
        total = len(vocab)
        return partial, total

    def create_tuples(self, docs = [], target = []):
        """
            Creates tuples with sample and its target

            Arguments:
                :docs (list[str]): list of texts
                :target (list[str]): list of targets
            Returns:
                list of tuples with sample and its target
        """
        return list(zip(docs, target))

    def create_vocab_from_tuples(self, tuples, is_lemma = True):
        """
            Creates vocabulary from list of tuples

            Arguments:
                :tuples (list[tuples]): list of tuples with sample and its target
                :is_lemma (bool): default is True
            Returns:
                dictionary of tokens with its count in an object collections.Counter
        """
        vocab = Counter()
        [self.add_doc_to_vocab(doc, vocab, is_lemma) for doc, target in tuples]
        return vocab

    def get_words_top(self, vocab, how_many):
        """
            Gets words top for each target

            Arguments:
                :vocab (collections.Counter): dictionary of tokens with its count
                :how_many (int): how many words in your top how_many list
            Returns:
                dictionary of the top how_many list
        """
        return {word for word, count in vocab.most_common(how_many)}

    def get_vocabs_cleaned(self, vocabs):
        """
            Cleans vocabs from common words among targets

            Arguments:
                :vocabs (dict): keys are targets, values are vocabularies for that target
            Returns:
                vocabs cleaned from common words among targets
        """
        targets = vocabs.keys()
        for target1 in targets:
            for target2 in targets:
                if target1 != target2:
                    common_set = set(vocabs[target1]).intersection(vocabs[target2])
                    for word in common_set:
                        del vocabs[target1][word]
                        del vocabs[target2][word]
        return vocabs

    def get_ngrams(self, degree = 2, doc = "", tokens = [], is_tuple = True, is_lemma = False):
        """
            Gets ngrams from doc or tokens

            Arguments:
                :degree (int): degree of ngrams, default is 2
                :doc (str): text, option if you pass tokens
                :tokens (list[str]): list of tokens, option if you pass doc
                :is_tuple (bool): default is True
                :is_lemma (bool): default is False
            Returns:
                list of tuples (n_grams) for that degree, or list of string (token)
        """
        ngrams = []
        if doc and not tokens:
            tokens = self.get_tokens_cleaned(doc, is_lemma)
        if tokens:
            ngrams = list(ng(tokens, degree))
        if is_tuple == False:
            n_grams = []
            for n_gram in ngrams:
                ngram = ' '.join(n_gram)
                n_grams.append(ngram)
            ngrams = n_grams
        return ngrams

    def get_ngrams_features(self, degree = 2, doc = "", tokens = [], is_lemma = False):
        """
            Gets ngrams features from doc or tokens
            Arguments:
                :degree (int): degree of ngrams, default is 2
                :doc (str): text, option if you pass tokens
                :tokens (list[str]): list of tokens, option if you pass doc
                :is_lemma (bool): default is False
            Returns:
                dictionary of ngrams extracted
        """
        features = {}
        ngrams = self.get_ngrams(degree = degree, doc = doc, tokens = tokens, is_tuple = False, is_lemma = is_lemma)
        for token in ngrams:
            features[token] = 1
        return features

    def create_ngrams_features_from_docs(self, docs, target, is_lemma = True, degree = 2):
        """
            Creates ngrams features from docs

            Arguments:
                :docs (list[str]): list of text
                :target (str): target name of the docs
                :is_lemma (bool): default is True
                :degree (int): degree of ngrams, default is 2
            Returns:
                list of tuples with features and relative target
        """
        return [(self.get_ngrams_features(degree = degree, doc = doc, is_lemma = is_lemma), target) for doc in docs]

    def create_ngrams_features_from_tuples(self, tuples, is_lemma = True, degree = 2):
        """
            Creates ngrams features from tuples

            Arguments:
                :tuples (list[tuples]): list of tuples with sample and its target
                :is_lemma (bool): default is True
                :degree (int): degree of ngrams, default is 2
            Returns:
                list of tuples with features and relative target
        """
        return [(self.get_ngrams_features(degree = degree, doc = doc, is_lemma = is_lemma), target) for (doc, target) in tuples]

    def get_features(self, doc, is_lemma = True, words_top = {}, degree = 0):
        """
            Gets features

            Arguments:
                :doc (str): text
                :is_lemma (bool): default is True
                :words_top (dict): dictionary of the words top
                :degree (int): degree of ngrams, default is 0
            Returns:
                dictionary of features extracted
        """
        tokens = self.get_tokens_cleaned(doc, is_lemma)
        features = {'words_top': 0}
        if degree > 0:
            features.update(self.get_ngrams_features(degree = degree, tokens = tokens, is_lemma = is_lemma))
        for token in tokens:
            if token in words_top:
                features['words_top'] += 1
        features.update(self.sia.polarity_scores(doc))
        return features

    def get_features_from_docs(self, docs, is_lemma = True, words_top = {}, degree = 0):
        """
            Gets features

            Arguments:
                :docs (list[str]): list of text
                :is_lemma (bool): default is True
                :words_top (dict): dictionary of the words top
                :degree (int): degree of ngrams, default is 0
            Returns:
                dictionary of features extracted
        """
        return [self.get_features(doc, is_lemma, words_top, degree) for doc in docs]

    def create_features_from_docs(self, docs, target, is_lemma = True, words_top = {}, degree = 0):
        """
            Creates features from docs

            Arguments:
                :docs (list[str]): list of text
                :target (str): target name of the docs
                :is_lemma (bool): default is True
                :words_top (dict): dictionary of the words top
                :degree (int): degree of ngrams, default is 0
            Returns:
                list of tuples with features and relative target
        """
        return [(self.get_features(doc, is_lemma, words_top, degree), target) for doc in docs]

    def create_features_from_tuples(self, tuples, is_lemma = True, words_top = {}, degree = 0):
        """
            Creates features from tuples

            Arguments:
                :tuples (list[tuples]): list of tuples with sample and its target
                :is_lemma (bool): default is True
                :words_top (dict): dictionary of the words top
                :degree (int): degree of ngrams, default is 0
            Returns:
                list of tuples with features and relative target
        """
        return [(self.get_features(doc, is_lemma, words_top, degree), target) for (doc, target) in tuples]

    def create_words_map(self, words):
        """
            Creates the map of words

            Arguments:
                :words (list[str]): words list
            Returns:
                string of all words
        """
        lower = [sentence.lower() for sentence in words]
        return ' '.join(map(str, lower))

    def create_words_cloud(self, words, is_test = False):
        """
            Creates the cloud of words

            Arguments:
                :words (str): words
                :is_test (bool): default is False
            Returns:
                only words cloud plot
        """
        words_cloud = WordCloud(width=1200, height=800, background_color="white").generate(words)
        if is_test == False:
            plt.imshow(words_cloud)
            plt.axis('off')
            plt.show()
        return words_cloud

    def vectorize_docs(self, docs, is_count = True, is_lemma = False, is_test = False):
        """
            Vectorizes docs

            Arguments:
                :docs (list[str]): list of texts
                :is_count (bool): default is True
                :is_lemma (bool): default is True
                :is_test (bool): default is False
            Returns:
                list of scipy.sparse.csr.csr_matrix, one for each doc
        """
        if type(docs) == list and type(docs[0]) == dict:
            vectorizer = DictVectorizer()
        elif is_count == True:
            vectorizer = CountVectorizer()
        else:
            vectorizer = TfidfVectorizer()
        if is_lemma == True:
            docs = [self.get_doc_cleaned(doc) for doc in docs]
        if is_test == True:            
            return self.vectorizer.transform(docs)
        else:
            self.vectorizer = vectorizer
            return self.vectorizer.fit_transform(docs)



import unittest
from smltk.preprocessing import Ntk
import nltk
from nltk.stem import WordNetLemmatizer
from collections import Counter
import scipy
import numpy as np
import pandas as pd

class TestNtk(unittest.TestCase, Ntk):
    ntk = None
    doc = 'Good case, Excellent value. I am agree. There is a mistake. Item Does Not Match Picture.'
    docs = []
    target = []
    tuples = []
    default_doc_filtered = 'good case excellent value agree mistake item match picture'
    without_stop_words_doc_filtered = 'good case excellent value i be agree there be a mistake item do not match picture'
    without_stop_words_with_min_length_doc_filtered = 'good case excellent value be agree there be mistake item do not match picture'
    without_stop_words_without_tag_map_doc_filtered = 'good case excellent value i am agree there is a mistake item doe not match picture'
    def __init__(self, *args, **kwargs):
        self.ntk = Ntk()
        self.ntk_no_stop_words = Ntk({'stop_words': []})
        self.docs = nltk.sent_tokenize(self.doc)
        self.target = [1, 1, 0, 0]
        self.tuples = self.ntk.create_tuples(self.docs, self.target)
        unittest.TestCase.__init__(self, *args, **kwargs)

    def test_get_tokens_cleaned(self):
        tokens = self.ntk_no_stop_words.get_tokens_cleaned(self.docs[0])
        self.assertEqual(tokens, ['good', 'case', 'excellent', 'value'])
        tokens = self.ntk_no_stop_words.get_tokens_cleaned(self.docs[1], is_lemma = True)
        self.assertEqual(tokens, ['i', 'be', 'agree'])
        tokens = self.ntk_no_stop_words.get_tokens_cleaned(self.docs[1], is_lemma = False)
        self.assertEqual(tokens, ['i', 'am', 'agree'])

        tokens = self.ntk.get_tokens_cleaned(self.docs[1], is_lemma = False)
        self.assertEqual(tokens, ['agree'])

    def test_get_doc_cleaned(self):
        self.assertEqual(self.ntk.get_doc_cleaned(self.doc), self.default_doc_filtered)
        self.assertEqual(self.ntk_no_stop_words.get_doc_cleaned(self.doc), self.without_stop_words_doc_filtered)

    def test_doc_cleaned(self):
        vocab_token = self.ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 9)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token, 2), (0, 9))

        vocab_lemma = self.ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 9)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_lemma), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_lemma, 2), (0, 9))

        features_token = self.ntk.create_features_from_tuples(self.tuples, False)
        self.assertEqual(features_token[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_token = self.ntk.create_features_from_docs(self.docs[:2], 1, False)
        features_token.extend(self.ntk.create_features_from_docs(self.docs[2:], 0, False))
        self.assertEqual(features_token[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_lemma = self.ntk.create_features_from_tuples(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_lemma = self.ntk.create_features_from_docs(self.docs[:2], 1, True)
        features_lemma.extend(self.ntk.create_features_from_docs(self.docs[2:], 0, True))
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        words_top = self.ntk.get_words_top(vocab_token, 4)
        features_lemma = self.ntk.create_features_from_tuples(self.tuples, True, words_top)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_lemma = self.ntk.create_features_from_docs(self.docs[:2], 1, True, words_top)
        features_lemma.extend(self.ntk.create_features_from_docs(self.docs[2:], 0, True, words_top))
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(self.ntk.get_doc_cleaned(doc, False))
        tuples = self.ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = self.ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token, 2), (0, 9))

        vocab_token = self.ntk.create_vocab_from_docs(docs_cleaned, False)
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token), (9, 9))
        self.assertEqual(self.ntk.get_stats_vocab(vocab_token, 2), (0, 9))

    def test_doc_cleaned_with_lemmatizer(self):
        ntk = Ntk({'lemmatizer': WordNetLemmatizer()})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.default_doc_filtered)

    def test_doc_cleaned_without_stop_words(self):
        vocab_token = self.ntk_no_stop_words.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 16)
        self.assertEqual(vocab_token['am'], 1)
        self.assertEqual(vocab_token['is'], 1)
        self.assertEqual(vocab_token['does'], 1)
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_token, 2), (0, 16))

        vocab_lemma = self.ntk_no_stop_words.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 15)
        self.assertEqual(vocab_lemma['be'], 2)
        self.assertEqual(vocab_lemma['do'], 1)
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_lemma), (15, 15))
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_lemma, 2), (1, 15))

        features_token = self.ntk_no_stop_words.create_features_from_tuples(self.tuples, False)
        self.assertEqual(features_token[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_lemma = self.ntk_no_stop_words.create_features_from_tuples(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        words_top = self.ntk.get_words_top(vocab_token, 4)
        features_lemma = self.ntk.create_features_from_tuples(self.tuples, True, words_top)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(self.ntk_no_stop_words.get_doc_cleaned(doc, False))
        tuples = self.ntk_no_stop_words.create_tuples(docs_cleaned, self.target)
        vocab_token = self.ntk_no_stop_words.create_vocab_from_tuples(tuples, False)
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_token, 2), (0, 16))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(self.ntk_no_stop_words.get_doc_cleaned(doc, True))
        tuples = self.ntk_no_stop_words.create_tuples(docs_cleaned, self.target)
        vocab_token = self.ntk_no_stop_words.create_vocab_from_tuples(tuples)
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_token), (15, 15))
        self.assertEqual(self.ntk_no_stop_words.get_stats_vocab(vocab_token, 2), (1, 15))

    def test_doc_cleaned_without_stop_words_with_min_length(self):
        ntk = Ntk({'stop_words': [], 'min_length': 2})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.without_stop_words_with_min_length_doc_filtered)

        vocab_token = ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 14)
        self.assertEqual(vocab_token['am'], 1)
        self.assertEqual(vocab_token['is'], 1)
        self.assertEqual(vocab_token['does'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (14, 14))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 14))

        vocab_lemma = ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 13)
        self.assertEqual(vocab_lemma['be'], 2)
        self.assertEqual(vocab_lemma['do'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma), (13, 13))
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma, 2), (1, 13))

        features_token = ntk.create_features_from_tuples(self.tuples, False)
        self.assertEqual(features_token[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_lemma = ntk.create_features_from_tuples(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        words_top = self.ntk.get_words_top(vocab_token, 4)
        features_lemma = self.ntk.create_features_from_tuples(self.tuples, True, words_top)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, False))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (14, 14))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 14))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, True))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (13, 13))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (1, 13))

    def test_doc_cleaned_without_stop_words_without_tag_map(self):
        ntk = Ntk({'stop_words': [], 'tag_map': {}})
        self.assertEqual(ntk.get_doc_cleaned(self.doc), self.without_stop_words_without_tag_map_doc_filtered)

        vocab_token = ntk.create_vocab_from_docs(self.docs, False)
        self.assertEqual(len(vocab_token), 16)
        self.assertEqual(vocab_token['does'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

        vocab_lemma = ntk.create_vocab_from_docs(self.docs, True)
        self.assertEqual(len(vocab_lemma), 16)
        self.assertEqual(vocab_lemma['doe'], 1)
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_lemma, 2), (0, 16))

        features_token = ntk.create_features_from_tuples(self.tuples, False)
        self.assertEqual(features_token[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        features_lemma = ntk.create_features_from_tuples(self.tuples, True)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        words_top = self.ntk.get_words_top(vocab_token, 4)
        features_lemma = self.ntk.create_features_from_tuples(self.tuples, True, words_top)
        self.assertEqual(features_lemma[1], ({'words_top': 0, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612}, 1))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, False))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples, False)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

        docs_cleaned = []
        for doc in self.docs:
            docs_cleaned.append(ntk.get_doc_cleaned(doc, True))
        tuples = ntk.create_tuples(docs_cleaned, self.target)
        vocab_token = ntk.create_vocab_from_tuples(tuples)
        self.assertEqual(ntk.get_stats_vocab(vocab_token), (16, 16))
        self.assertEqual(ntk.get_stats_vocab(vocab_token, 2), (0, 16))

    def test_get_words_top(self):
        vocabs = Counter({'one': 10, 'two': 8, 'three': 4, 'four': 2, 'five': 0, 'six': 1})
        vocabs_cleaned = self.ntk.get_words_top(vocabs, 4)
        self.assertEqual(vocabs_cleaned, {'one', 'two', 'three', 'four'})

    def test_get_vocabs_cleaned(self):
        vocabs = {'one': Counter({'uno':1, 'uno':2, 'hm':3}), 'two': Counter({'due':2, 'dos':3, 'hm':4})}
        vocabs_cleaned = self.ntk.get_vocabs_cleaned(vocabs)
        self.assertEqual(vocabs_cleaned, {'one': Counter({'uno':1, 'uno':2}), 'two': Counter({'due':2, 'dos':3})})
        self.assertEqual(vocabs_cleaned, {'one': {'uno':1, 'uno':2}, 'two': {'due':2, 'dos':3}})
        vocabs = {'one': {'uno':1, 'uno':2, 'hm':3}, 'two': {'due':2, 'dos':3, 'hm':4}}
        vocabs_cleaned = self.ntk.get_vocabs_cleaned(vocabs)
        self.assertEqual(vocabs_cleaned, {'one': Counter({'uno':1, 'uno':2}), 'two': Counter({'due':2, 'dos':3})})
        self.assertEqual(vocabs_cleaned, {'one': {'uno':1, 'uno':2}, 'two': {'due':2, 'dos':3}})

    def test_get_ngrams(self):
        tokens = self.ntk.get_tokens_cleaned(self.doc)
        ngrams = self.ntk.get_ngrams()
        self.assertEqual(ngrams, [])
        ngrams = self.ntk.get_ngrams(degree = 1, tokens = tokens)
        self.assertEqual(type(ngrams[0]), tuple)
        ngrams = self.ntk.get_ngrams(degree = 1, doc = self.doc)
        self.assertEqual(type(ngrams[0]), tuple)
        ngrams = self.ntk.get_ngrams(degree = 1, doc = self.doc, is_tuple = False)
        np.testing.assert_array_equal(ngrams, tokens)

    def test_get_ngrams_features(self):
        tokens = self.ntk_no_stop_words.get_tokens_cleaned(self.doc, is_lemma = False)
        features = self.ntk_no_stop_words.get_ngrams_features()
        self.assertEqual(features, {})
        tokens_features = self.ntk_no_stop_words.get_ngrams_features(degree = 1, tokens = tokens)
        self.assertEqual(tokens_features, {'good': 1, 'case': 1, 'excellent': 1, 'value': 1, 'i': 1, 'am': 1, 'agree': 1, 'there': 1, 'is': 1, 'a': 1, 'mistake': 1, 'item': 1, 'does': 1, 'not': 1, 'match': 1, 'picture': 1})
        doc_features = self.ntk_no_stop_words.get_ngrams_features(degree = 1, doc = self.doc)
        self.assertEqual(doc_features, tokens_features)
        features = self.ntk_no_stop_words.get_ngrams_features(degree = 1, doc = self.doc, is_lemma = True)
        self.assertEqual(features, {'good': 1, 'case': 1, 'excellent': 1, 'value': 1, 'i': 1, 'be': 1, 'agree': 1, 'there': 1, 'a': 1, 'mistake': 1, 'item': 1, 'do': 1, 'not': 1, 'match': 1, 'picture': 1})

    def test_create_ngrams_features_from_docs(self):
        features = self.ntk.create_ngrams_features_from_docs(self.docs, True)
        self.assertEqual(type(features[0]), tuple)
        self.assertEqual(features[0][1], True)
        self.assertEqual(features[1][0], {})

    def test_create_ngrams_features_from_tuples(self):
        features = self.ntk.create_ngrams_features_from_tuples(self.tuples, True)
        self.assertEqual(type(features[0]), tuple)
        self.assertEqual(features[0][1], True)
        self.assertEqual(features[1][0], {})

    def test_get_features(self):
        features = self.ntk.get_features(self.docs[0])
        self.assertEqual(features, {'words_top': 0, 'neg': 0.0, 'neu': 0.1, 'pos': 0.9, 'compound': 0.8402})
        features = self.ntk_no_stop_words.get_features(self.docs[1], is_lemma = True, degree = 1)
        self.assertEqual(features, {'words_top': 0, 'i': 1, 'be': 1, 'agree': 1, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612})
        features = self.ntk_no_stop_words.get_features(self.docs[1], is_lemma = False, degree = 1)
        self.assertEqual(features, {'words_top': 0, 'i': 1, 'am': 1, 'agree': 1, 'neg': 0.0, 'neu': 0.286, 'pos': 0.714, 'compound': 0.3612})

    def create_words_map(self):
        tokens = self.ntk.get_tokens_cleaned(self.doc)
        return self.ntk.create_words_map(tokens)

    def test_create_words_map(self):
        words_map = self.create_words_map()
        self.assertEqual(words_map, self.default_doc_filtered)

    def test_create_words_cloud(self):
        words_map = self.create_words_map()
        words_cloud = self.ntk.create_words_cloud(words_map, True)
        np.testing.assert_array_equal(words_cloud.words_, {'good': 1.0, 'case': 1.0, 'excellent': 1.0, 'value': 1.0, 'agree': 1.0, 'mistake': 1.0, 'item': 1.0, 'match': 1.0, 'picture': 1.0})

    def test_vectorize_docs(self):
        data = pd.DataFrame(self.docs, columns=['text'])
        features = self.ntk.get_features_from_docs(self.docs)
        features_df = pd.DataFrame.from_dict(features, orient='columns')
        data = pd.concat([data, features_df], axis='columns')
        features_mix = data[['text', 'words_top', 'neg', 'neu', 'pos', 'compound']].to_dict(orient="records")

        X_train_dict = self.ntk.vectorize_docs(features_mix)
        self.assertEqual(type(X_train_dict[0]), scipy.sparse.csr.csr_matrix)
        vectorizer_dict = self.ntk.vectorizer
        self.assertEqual(len(vectorizer_dict.vocabulary_), 9)
        X_test_dict = self.ntk.vectorize_docs(features_mix, is_test = True)
        self.assertEqual(type(X_test_dict[0]), scipy.sparse.csr.csr_matrix)

        X_train_count = self.ntk.vectorize_docs(self.docs)
        self.assertEqual(type(X_train_count[0]), scipy.sparse.csr.csr_matrix)
        vectorizer_count = self.ntk.vectorizer
        self.assertEqual(len(vectorizer_count.vocabulary_), 14)
        X_test_count = self.ntk.vectorize_docs(self.docs, is_test = True)
        self.assertEqual(type(X_test_count[0]), scipy.sparse.csr.csr_matrix)

        X_train_tfidf = self.ntk.vectorize_docs(self.docs, False)
        self.assertEqual(type(X_train_tfidf[0]), scipy.sparse.csr.csr_matrix)
        vectorizer_tfidf = self.ntk.vectorizer
        self.assertEqual(len(vectorizer_tfidf.vocabulary_), 14)
        X_test_tfidf = self.ntk.vectorize_docs(self.docs, is_test = True)
        self.assertEqual(type(X_test_tfidf[0]), scipy.sparse.csr.csr_matrix)

        X_train_count = self.ntk.vectorize_docs(self.docs, is_lemma = True)
        self.assertEqual(type(X_train_count[0]), scipy.sparse.csr.csr_matrix)
        vectorizer_lemma = self.ntk.vectorizer
        self.assertEqual(len(vectorizer_lemma.vocabulary_), 9)

        X_train_tfidf = self.ntk.vectorize_docs(self.docs, is_count = False, is_lemma = True)
        self.assertEqual(type(X_train_tfidf[0]), scipy.sparse.csr.csr_matrix)
        vectorizer_lemma = self.ntk.vectorizer
        self.assertEqual(len(vectorizer_lemma.vocabulary_), 9)

if __name__ == '__main__':
    unittest.main()


import unittest
from smltk.metrics import Metrics
from smltk.preprocessing import Ntk
import nltk
import numpy as np
import pandas as pd
from sklearn.datasets import load_wine
from sklearn.model_selection import train_test_split
from sklearn.linear_model import SGDClassifier
import os

class TestMetrics(unittest.TestCase, Metrics):
    mtr = None
    ntk = None
    doc = 'Good case, Excellent value. I am agree. There is a mistake. Item Does Not Match Picture.'
    docs = []
    target = []
    tuples = []
    def __init__(self, *args, **kwargs):
        self.mtr = Metrics()
        self.ntk = Ntk()
        self.docs = nltk.sent_tokenize(self.doc)
        self.target = [1, 1, 0, 0]
        self.tuples = self.ntk.create_tuples(self.docs, self.target)
        unittest.TestCase.__init__(self, *args, **kwargs)

    def training(self):
        features_lemma = self.ntk.create_features_from_tuples(self.tuples, True)
        classifier = nltk.NaiveBayesClassifier.train(features_lemma)
        return classifier, features_lemma

    def prediction(self):
        classifier, features_lemma = self.training()
        y_test, y_pred = self.mtr.prediction(classifier, 'classify', features_lemma)
        return y_test, y_pred

    def test_split_tuples(self):
        classifier, features_lemma = self.training()
        X_test, y_test = self.split_tuples(features_lemma)
        self.assertEqual(y_test, [1, 1, 0, 0])

    def test_prediction(self):
        classifier, features_lemma = self.training()
        y_test, y_pred = self.mtr.prediction(classifier, 'classify', features_lemma)
        self.assertEqual(y_test, y_pred)
        y_test, y_pred = self.mtr.prediction(classifier, 'classify', [])
        self.assertEqual(y_test, y_pred)
        np.testing.assert_array_equal(y_pred, [])
        X_test = pd.Series(features_lemma)
        y_test, y_pred = self.mtr.prediction(classifier, 'classify', X_test)
        self.assertEqual(y_test, y_pred)

    def test_create_confusion_matrix(self):
        y_test, y_pred = self.prediction()
        np.testing.assert_array_equal(self.mtr.create_confusion_matrix(y_test, y_pred, True), [[2, 0], [0, 2]])

    def test_is_binary_classification(self):
        y_test = [1, 0, 1]
        y_pred = [0, 1, 0]
        self.assertTrue(self.mtr.is_binary_classification(y_test, y_pred))
        y_pred = [0, 1, 2]
        self.assertFalse(self.mtr.is_binary_classification(y_test, y_pred))

    def test_clean_binary_classification(self):
        y_test = [1, 0, 1]
        y_pred = [0, 1, 0]
        y_test_cleaned, y_pred_cleaned = self.mtr.clean_binary_classification(y_test, y_pred)
        self.assertEqual(y_test_cleaned[0], 1)
        self.assertEqual(y_test_cleaned[1], 0)
        self.assertEqual(y_pred_cleaned[0], 0)
        self.assertEqual(y_pred_cleaned[1], 1)
        y_test = ['pos', 'neg', 'pos']
        y_pred = ['neg', 'pos', 'neg']
        y_test_cleaned, y_pred_cleaned = self.mtr.clean_binary_classification(y_test, y_pred)
        self.assertEqual(y_test_cleaned[0], 1)
        self.assertEqual(y_test_cleaned[1], 0)
        self.assertEqual(y_pred_cleaned[0], 0)
        self.assertEqual(y_pred_cleaned[1], 1)

    def test_get_classification_metrics(self):
        classifier, features_lemma = self.training()
        y_test, y_pred = self.mtr.prediction(classifier, 'classify', features_lemma)
        X_test, y_test = self.mtr.split_tuples(features_lemma)
        params = {
            "model": classifier,
            "X_train": np.array(X_test),
            "y_train": np.array(y_test),
            "X_test": np.array(X_test),
            "y_test": np.array(y_test),
            "y_pred": y_pred
        }
        metrics = self.mtr.get_classification_metrics(params)
        self.assertEqual(metrics['Loss'], 0)
        self.assertEqual(metrics['Accuracy'], 1.0)
        self.assertEqual(metrics['MCC'], 1.0)
        self.assertEqual(metrics['ROC_AUC'], 1.0)
        np.testing.assert_array_equal(metrics['Precision'], [1., 1.])
        np.testing.assert_array_equal(metrics['Support'], [2, 2])
        params = {
            "y_test": np.array(y_test),
            "y_pred": y_pred
        }
        metrics = self.mtr.get_classification_metrics(params)
        self.assertEqual(metrics['Loss'], 0)
        self.assertEqual(metrics['Accuracy'], 1.0)
        self.assertEqual(metrics['MCC'], 1.0)
        self.assertEqual(metrics['ROC_AUC'], 1.0)
        np.testing.assert_array_equal(metrics['Precision'], [1., 1.])
        np.testing.assert_array_equal(metrics['Support'], [2, 2])

        data = load_wine()
        X_train, X_test, y_train, y_test = train_test_split(data.data, data.target, test_size=0.2, random_state=5)
        model = SGDClassifier(random_state=3)
        model.fit(X_train, y_train)
        y_pred = model.predict(X_test)
        params = {
            "model": model,
            "X_train": np.array(X_train),
            "y_train": np.array(y_train),
            "X_test": np.array(X_test),
            "y_test": np.array(y_test),
            "y_pred": y_pred
        }
        metrics = self.mtr.get_classification_metrics(params)
        self.assertEqual(metrics['Loss'], 0.7443055555555557)
        self.assertEqual(metrics['Accuracy'], 0.6666666666666666)
        self.assertEqual(metrics['MCC'], 0.4802259242337604)
        np.testing.assert_array_equal(metrics['Precision'][2], 0.4)
        np.testing.assert_array_equal(metrics['Support'][2], 8)

    def test_manage_model(self):
        filename = '/tmp/save.model'
        classifier, features_lemma = self.training()
        y_test1, y_pred1 = self.mtr.prediction(classifier, 'classify', features_lemma)
        self.mtr.save_model(classifier, filename)
        self.assertFalse(os.stat(filename).st_size == 0)
        model = self.mtr.resume_model(filename)
        y_test2, y_pred2 = self.mtr.prediction(model, 'classify', features_lemma)
        self.assertEqual(y_test1, y_test2)
        self.assertEqual(y_pred1, y_pred2)

if __name__ == '__main__':
    unittest.main()